@@ -655,13 +655,17 @@ static bool dhcp_message_compare(const uint8_t *expected, size_t expected_len,
 	return r;
 }
 
+static bool client_send_called = false;
+
 static int fake_transport_send(struct dhcp_transport *transport,
 				const struct sockaddr_in *dest,
 				const void *data, size_t len)
 {
 	assert(len <= sizeof(client_packet));
+	assert(!client_send_called);
 	memcpy(client_packet, data, len);
 	client_packet_len = len;
+	client_send_called = true;
 
 	return len;
 }
@@ -673,8 +677,10 @@ static int fake_transport_l2_send(struct dhcp_transport *transprot,
 					const void *data, size_t len)
 {
 	assert(len <= sizeof(client_packet));
+	assert(!client_send_called);
 	memcpy(client_packet, data, len);
 	client_packet_len = len;
+	client_send_called = true;
 
 	return len;
 }
@@ -710,16 +716,21 @@ static void test_discover(const void *data)
 
 	assert(l_dhcp_client_start(client));
 
+	assert(client_send_called);
+	client_send_called = false;
 	assert(dhcp_message_compare(discover_data_1, sizeof(discover_data_1),
 					client_packet, client_packet_len));
 
-	transport->rx_cb(offer_data_1, sizeof(offer_data_1), client);
+	transport->rx_cb(offer_data_1, sizeof(offer_data_1), client, NULL);
 
+	assert(client_send_called);
+	client_send_called = false;
 	assert(dhcp_message_compare(request_data_1, sizeof(request_data_1),
 					client_packet, client_packet_len));
 
 	event_handler_called = false;
-	transport->rx_cb(ack_data_1, sizeof(ack_data_1), client);
+	transport->rx_cb(ack_data_1, sizeof(ack_data_1), client, NULL);
+	assert(!client_send_called);
 	assert(event_handler_called);
 
 	lease = l_dhcp_client_get_lease(client);
@@ -736,6 +747,8 @@ static void test_discover(const void *data)
 	assert(lease->t2 == 0x00012750);
 
 	l_dhcp_client_destroy(client);
+	assert(client_send_called);
+	client_send_called = false;
 }
 
 static bool l2_send_called = false;
@@ -749,6 +762,7 @@ static int fake_transport_server_l2_send(struct dhcp_transport *s,
 					const void *data, size_t len)
 {
 	assert(len <= sizeof(server_packet));
+	assert(!l2_send_called);
 	memcpy(server_packet, data, len);
 	server_packet_len = len;
 
@@ -774,9 +788,11 @@ static void server_event(struct l_dhcp_server *server,
 {
 	switch (event) {
 	case L_DHCP_SERVER_EVENT_NEW_LEASE:
+		assert(!new_client);
 		new_client = l_dhcp_lease_get_address(lease);
 		break;
 	case L_DHCP_SERVER_EVENT_LEASE_EXPIRED:
+		assert(!expired_client);
 		expired_client = l_dhcp_lease_get_address(lease);
 		break;
 	}
@@ -806,35 +822,46 @@ static struct l_dhcp_client *client_init(const uint8_t *mac)
 	return client;
 }
 
-static bool client_connect(struct l_dhcp_client *client,
-				struct l_dhcp_server *server)
+static void client_connect(struct l_dhcp_client *client,
+				struct l_dhcp_server *server, bool rapid_commit)
 {
 	struct dhcp_transport *srv_transport =
 					_dhcp_server_get_transport(server);
 	struct dhcp_transport *cli_transport =
 					_dhcp_client_get_transport(client);
-	if (!l_dhcp_client_start(client))
-		return false;
+	uint8_t cli_addr[ETH_ALEN];
+	const struct dhcp_message *msg = (struct dhcp_message *) client_packet;
+
+	assert(l_dhcp_client_start(client));
+	assert(client_send_called);
+	client_send_called = false;
+	memcpy(cli_addr, msg->chaddr, ETH_ALEN);
 
 	/* RX DISCOVER */
-	srv_transport->rx_cb(client_packet, client_packet_len, server);
+	srv_transport->rx_cb(client_packet, client_packet_len, server,
+				cli_addr);
 	assert(l2_send_called);
 	l2_send_called = false;
 
-	/* RX OFFER */
-	cli_transport->rx_cb(server_packet, server_packet_len, client);
+	if (!rapid_commit) {
+		/* RX OFFER */
+		cli_transport->rx_cb(server_packet, server_packet_len, client,
+					NULL);
+		assert(client_send_called);
+		client_send_called = false;
 
-	/* RX REQUEST */
-	srv_transport->rx_cb(client_packet, client_packet_len, server);
-	assert(l2_send_called);
-	l2_send_called = false;
+		/* RX REQUEST */
+		srv_transport->rx_cb(client_packet, client_packet_len, server,
+					cli_addr);
+		assert(l2_send_called);
+		l2_send_called = false;
+	}
 
 	/* RX ACK */
-	cli_transport->rx_cb(server_packet, server_packet_len, client);
+	cli_transport->rx_cb(server_packet, server_packet_len, client, NULL);
+	assert(!client_send_called);
 
 	assert(event_handler_called);
-
-	return true;
 }
 
 static struct l_dhcp_server *server_init()
@@ -866,6 +893,7 @@ static struct l_dhcp_server *server_init()
 
 static void test_complete_run(const void *data)
 {
+	bool rapid_commit = L_PTR_TO_UINT(data);
 	static const uint8_t addr1[6] = { 0x01, 0x02, 0x03, 0x04, 0x05, 0x06 };
 	static const uint8_t addr2[6] = { 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c };
 	struct dhcp_transport *srv_transport;
@@ -883,6 +911,7 @@ static void test_complete_run(const void *data)
 	char **dns_list;
 
 	server = server_init();
+	l_dhcp_server_set_enable_rapid_commit(server, rapid_commit);
 
 	assert(l_dhcp_server_set_ip_range(server, "192.168.1.2",
 						"192.168.1.100"));
@@ -891,7 +920,7 @@ static void test_complete_run(const void *data)
 
 	client1 = client_init(addr1);
 
-	client_connect(client1, server);
+	client_connect(client1, server, rapid_commit);
 
 	cli_lease = l_dhcp_client_get_lease(client1);
 	assert(cli_lease);
@@ -916,6 +945,8 @@ static void test_complete_run(const void *data)
 	assert(!strcmp(tmp_addr, "255.255.255.0"));
 	l_free(tmp_addr);
 
+	assert(l_dhcp_lease_get_prefix_length(cli_lease) == 24);
+
 	dns_list = l_dhcp_lease_get_dns(cli_lease);
 	assert(dns_list && dns_list[0] && dns_list[1]);
 	assert(!strcmp(dns_list[0], "192.168.1.1"));
@@ -924,7 +955,7 @@ static void test_complete_run(const void *data)
 
 	client2 = client_init(addr2);
 
-	client_connect(client2, server);
+	client_connect(client2, server, rapid_commit);
 
 	cli_lease = l_dhcp_client_get_lease(client2);
 	assert(cli_lease);
@@ -941,22 +972,27 @@ static void test_complete_run(const void *data)
 	l_free(cli_addr);
 
 	l_dhcp_client_stop(client1);
-
-	srv_transport->rx_cb(client_packet, client_packet_len, server);
+	assert(client_send_called);
+	client_send_called = false;
+	srv_transport->rx_cb(client_packet, client_packet_len, server, addr1);
 	assert(expired_client);
 	assert(!strcmp(expired_client, "192.168.1.2"));
 	l_free(expired_client);
 	expired_client = NULL;
 
 	l_dhcp_client_stop(client2);
-	srv_transport->rx_cb(client_packet, client_packet_len, server);
+	assert(client_send_called);
+	client_send_called = false;
+	srv_transport->rx_cb(client_packet, client_packet_len, server, addr2);
 	assert(expired_client);
 	assert(!strcmp(expired_client, "192.168.1.3"));
 	l_free(expired_client);
 	expired_client = NULL;
 
 	l_dhcp_client_destroy(client1);
+	assert(!client_send_called);
 	l_dhcp_client_destroy(client2);
+	assert(!client_send_called);
 
 	l_dhcp_server_stop(server);
 	l_dhcp_server_destroy(server);
@@ -975,6 +1011,7 @@ static void test_expired_ip_reuse(const void *data)
 
 	l_dhcp_server_set_ip_range(server, "192.168.1.2", "192.168.1.11");
 	_dhcp_server_set_max_expired_clients(server, 10);
+	l_dhcp_server_set_enable_rapid_commit(server, false);
 
 	/*
 	 * Connect and release 10 clients, this should max out the expired
@@ -987,19 +1024,22 @@ static void test_expired_ip_reuse(const void *data)
 
 		addr[5] = i;
 		client = client_init(addr);
-		client_connect(client, server);
+		client_connect(client, server, false);
 		l_free(new_client);
 		new_client = NULL;
 
 		l_dhcp_client_destroy(client);
-		srv_transport->rx_cb(client_packet, client_packet_len, server);
+		assert(client_send_called);
+		client_send_called = false;
+		srv_transport->rx_cb(client_packet, client_packet_len, server,
+					addr);
 		l_free(expired_client);
 		expired_client = NULL;
 	}
 
 	addr[5] = i + 1;
 	client_new = client_init(addr);
-	client_connect(client_new, server);
+	client_connect(client_new, server, false);
 	l_free(new_client);
 	new_client = NULL;
 
@@ -1010,6 +1050,8 @@ static void test_expired_ip_reuse(const void *data)
 	l_free(cli_addr);
 
 	l_dhcp_client_destroy(client_new);
+	assert(client_send_called);
+	client_send_called = false;
 	l_dhcp_server_destroy(server);
 }
 
@@ -1036,7 +1078,8 @@ int main(int argc, char *argv[])
 
 	l_test_add("discover", test_discover, NULL);
 
-	l_test_add("complete run", test_complete_run, NULL);
+	l_test_add("complete run", test_complete_run, L_UINT_TO_PTR(false));
+	l_test_add("rapid commit", test_complete_run, L_UINT_TO_PTR(true));
 	l_test_add("expired IP reuse", test_expired_ip_reuse, NULL);
 
 	return l_test_run();
